@@ -42,32 +42,38 @@ def ModifiedCausalForest(
         datpfad=None, descriptive_stats=True, dpi=500,
         fs_other_sample_share=0.2, fs_other_sample=True, fs_rf_threshold=0,
         fs_yes=None, fontsize=2, forest_files=None, gatet_flag=False,
-        gmate_no_evaluation_points=50, gmate_sample_share=None, indata=None,
-        knn_flag=False, knn_min_k=10, knn_const=1, l_centering=False,
-        l_centering_share=0.25, l_centering_cv_k=5,
-        l_centering_new_sample=False, m_min_share=-1, m_max_share=-1, m_grid=2,
-        m_random_poisson=True, match_nn_prog_score=True, max_cats_z_vars=None,
-        max_weight_share=0.05, mce_vart=1, min_dummy_obs=10,  mp_parallel=None,
-        mp_ray_objstore_multiplier=None, mp_with_ray=True, mp_vim_type=None,
-        mp_weights_tree_batch=None,  mp_weights_type=1, n_min_grid=1,
-        n_min_min=-1, n_min_max=-1, nn_main_diag_only=False, no_filled_plot=20,
-        nw_kern_flag=1, nw_bandw=None, outfiletext=None, outpfad=None,
-        output_type=2, panel_data=False, panel_in_rf=True, preddata=None,
-        p_diff_penalty=None, post_est_stats=True, post_plots=True,
-        post_kmeans_yes=True, post_kmeans_max_tries=1000,
+        gmate_no_evaluation_points=50, gmate_sample_share=None, iate_flag=True,
+        iate_se_flag=True, indata=None, knn_flag=True, knn_min_k=10,
+        knn_const=1, l_centering=True, l_centering_share=0.25,
+        l_centering_cv_k=5, l_centering_new_sample=False, m_min_share=-1,
+        m_max_share=-1, m_grid=2, m_random_poisson=True,
+        match_nn_prog_score=True, max_cats_z_vars=None, max_weight_share=0.05,
+        mce_vart=1, min_dummy_obs=10,  mp_parallel=None, mp_with_ray=True,
+        mp_vim_type=None, mp_weights_tree_batch=None,  mp_weights_type=1,
+        n_min_grid=1, n_min_min=-1, n_min_max=-1, nn_main_diag_only=False,
+        no_filled_plot=20, nw_kern_flag=1, nw_bandw=None, outfiletext=None,
+        outpfad=None, output_type=2, panel_data=False, panel_in_rf=True,
+        preddata=None, p_diff_penalty=None, post_est_stats=True,
+        post_plots=True, post_kmeans_yes=True, post_kmeans_max_tries=1000,
         post_kmeans_no_of_groups=None, post_kmeans_replications=10,
-        post_random_forest_vi=True, predict_mcf=True,
-        random_thresholds=20, relative_to_first_group_only=True,
-        save_forest=False, screen_covariates=True, share_forest_sample=0.5,
-        show_plots=True, smooth_gates=True, smooth_gates_bandwidth=1,
-        smooth_gates_no_evaluation_points=50, stop_empty=25,
-        subsample_factor=None, support_check=1, support_min_p=None,
-        support_quantil=1, support_max_del_train=0.5,
-        train_mcf=True, variable_importance_oob=False,
-        verbose=True, weight_as_sparse=True, weighted=False,
-        se_boot_ate=False, se_boot_gate=False, se_boot_iate=False,
-        _smaller_sample=0, _with_output=True, _max_cats_cont_vars=None,
-        _max_save_values=50, _seed_sample_split=67567885):
+        post_random_forest_vi=True, predict_mcf=True, random_thresholds=None,
+        reduce_prediction=False, reduce_prediction_share=0.5,
+        reduce_split_sample=False, reduce_split_sample_pred_share=0.5,
+        reduce_training=False, reduce_training_share=0.5,
+        relative_to_first_group_only=True, reduce_largest_group_train=False,
+        reduce_largest_group_train_share=0.5, save_forest=False,
+        screen_covariates=True, se_boot_ate=False, se_boot_gate=False,
+        se_boot_iate=False, share_forest_sample=0.5, show_plots=True,
+        smooth_gates=True, smooth_gates_bandwidth=1,
+        smooth_gates_no_evaluation_points=50, stop_empty=1,
+        subsample_factor_forest=None, subsample_factor_eval=False,
+        support_check=1, support_min_p=None, support_quantil=1,
+        support_max_del_train=0.5, train_mcf=True,
+        variable_importance_oob=False, verbose=True, weight_as_sparse=True,
+        weighted=False, _smaller_sample=0, _with_output=True,
+        _max_cats_cont_vars=None, _max_save_values=50,
+        _seed_sample_split=67567885, _mp_ray_del=None,
+        _mp_ray_shutdown=None, _mp_ray_objstore_multiplier=None):
     """Compute the honest causal/random forest (based on mcf)."""
     freeze_support()
     time1 = time.time()
@@ -78,8 +84,7 @@ def ModifiedCausalForest(
     _max_elements_per_split = 100 * 10e5  # reduce if breakdown (weights)
     _load_old_forest = False  # Useful for testing to save time
     _no_ray_in_forest_building = False
-    # Problem in Ray 1.2.0 when large forests are build; check if still there
-    # Ray 2.0.0 once available - concurrent futures even more problematic!
+    _l_centering_uncenter=False
 
 # Collect vars in a dictionary
     variable_dict = mcf_init.make_user_variable(
@@ -92,10 +97,12 @@ def ModifiedCausalForest(
 # use smaller random sample (usually for testing purposes)
     if 0 < _smaller_sample < 1:
         if train_mcf:
-            gp.randomsample(datpfad, indata + '.csv', 'smaller_idata.csv',
+            gp.randomsample(datpfad, indata + '.csv', 'smaller_indata.csv',
                             _smaller_sample, True)
+        if preddata is None:
+            preddata = 'smaller_indata'
         if predict_mcf and preddata != indata:
-            gp.randomsample(datpfad, preddata + '.csv', 'smaller_pdata.csv',
+            gp.randomsample(datpfad, preddata + '.csv', 'smaller_preddata.csv',
                             _smaller_sample, True)
             preddata = 'smaller_preddata'
         else:
@@ -108,10 +115,11 @@ def ModifiedCausalForest(
         n_min_grid, check_perfectcorr, n_min_min, clean_data_flag,
         min_dummy_obs, mce_vart, p_diff_penalty, boot, n_min_max,
         support_min_p, weighted, support_check, support_quantil,
-        subsample_factor, m_min_share, m_grid, stop_empty,
-        m_random_poisson, alpha_reg_min, alpha_reg_max, alpha_reg_grid,
-        random_thresholds, knn_min_k, share_forest_sample, descriptive_stats,
-        m_max_share, max_cats_z_vars, variable_importance_oob,
+        subsample_factor_forest, subsample_factor_eval, m_min_share, m_grid,
+        stop_empty, m_random_poisson, alpha_reg_min, alpha_reg_max,
+        alpha_reg_grid, random_thresholds, knn_min_k, share_forest_sample,
+        descriptive_stats, m_max_share, max_cats_z_vars,
+        variable_importance_oob,
         balancing_test, choice_based_sampling, knn_const, choice_based_weights,
         nw_kern_flag, post_kmeans_max_tries, cond_var_flag, knn_flag, nw_bandw,
         panel_data, _max_cats_cont_vars, cluster_std, fs_yes,
@@ -127,9 +135,14 @@ def ModifiedCausalForest(
         smooth_gates_bandwidth, smooth_gates_no_evaluation_points, show_plots,
         weight_as_sparse, mp_weights_type, mp_weights_tree_batch,
         _boot_by_boot, _obs_by_obs, _max_elements_per_split, mp_with_ray,
-        mp_ray_objstore_multiplier, verbose, _no_ray_in_forest_building,
+        _mp_ray_objstore_multiplier, verbose, _no_ray_in_forest_building,
         predict_mcf, train_mcf, forest_files, match_nn_prog_score,
-        se_boot_ate, se_boot_gate, se_boot_iate, support_max_del_train)
+        se_boot_ate, se_boot_gate, se_boot_iate, support_max_del_train,
+        _mp_ray_del, _mp_ray_shutdown,
+        reduce_split_sample, reduce_split_sample_pred_share, reduce_training,
+        reduce_training_share, reduce_prediction, reduce_prediction_share,
+        reduce_largest_group_train, reduce_largest_group_train_share,
+        iate_flag, iate_se_flag, _l_centering_uncenter)
 # Set defaults for many control variables of the MCF & define variables
 
     c_dict, v_dict, text_to_print = mcf_init.get_controls(controls_dict,
@@ -144,6 +157,15 @@ def ModifiedCausalForest(
             else:
                 outfiletext = open(c_dict['outfiletext'], 'w')
                 sys.stdout = outfiletext
+    if c_dict['reduce_split_sample']:
+        outdatei1 = c_dict['indata'][:-4] + 'estredu.csv'
+        outdatei2 = c_dict['preddata'][:-4] + 'predredu.csv'
+        c_dict['indata'], c_dict['preddata'], _ = gp.sample_split_2_3(
+            c_dict['indata'], outdatei1,
+            1-c_dict['reduce_split_sample_pred_share'], outdatei2,
+            c_dict['reduce_split_sample_pred_share'],
+            random_seed=_seed_sample_split,
+            with_output=c_dict['with_output'])
     if c_dict['with_output'] and c_dict['verbose']:
         if c_dict['train_mcf']:
             if text_to_print is not None:
@@ -152,8 +174,11 @@ def ModifiedCausalForest(
             gp.print_dic(c_dict)
             print('\nVariables used:')
             gp.print_dic(v_dict)
+            print('-' * 80)
+            print('Estimator used: ', c_dict['estimator_str'])
             print('\nValue of penalty multiplier:',
                   '{:8.3f}'.format(c_dict['mtot_p_diff_penalty']))
+            print('-' * 80)
             if c_dict['desc_stat']:
                 gp.print_descriptive_stats_file(c_dict['indata'], 'all',
                                                 c_dict['print_to_file'])
@@ -185,6 +210,7 @@ def ModifiedCausalForest(
         z_new_name_dict = loaded_tuple[10]
         z_new_dic_dict = loaded_tuple[11]
         c_dict['max_cats_cont_vars'] = loaded_tuple[12]
+        lc_forest = loaded_tuple[13]
     else:
         d_in_values = None
         no_val_dict = None
@@ -195,6 +221,8 @@ def ModifiedCausalForest(
         common_support_list = None
         z_new_name_dict = None
         z_new_dic_dict = None
+        lc_forest = None
+
 # Prepare data: Add and recode variables for GATES (Z)
 #               Recode categorical variables to prime numbers, cont. vars
     (v_dict, var_x_type, var_x_values, c_dict, indata_with_z, predata_with_z,
@@ -204,6 +232,17 @@ def ModifiedCausalForest(
          v_dict, c_dict, False, d_in_values, no_val_dict, q_inv_dict,
          q_inv_cr_dict, prime_values_dict, unique_val_dict,
          z_new_name_dict, z_new_dic_dict)
+
+    if c_dict['with_output'] and c_dict['verbose']:
+        print()
+        print('Dictionary for recoding of categorical variables into primes')
+        key_list_unique_val_dict = list(unique_val_dict)
+        key_list_prime_values_dict = list(prime_values_dict)
+        for key_nr, key in enumerate(key_list_unique_val_dict):
+            print(key)
+            print(unique_val_dict[key])
+            print(key_list_prime_values_dict[key_nr])  # order may differ!
+            print(prime_values_dict[key_list_prime_values_dict[key_nr]])
     # Remove missing and keep only variables needed for further analysis
     if c_dict['clean_data_flag']:
         if c_dict['train_mcf']:
@@ -217,7 +256,7 @@ def ModifiedCausalForest(
                 c_dict['desc_stat'], c_dict['print_to_file'])
         if indata_with_z != predata_with_z and c_dict['pred_mcf']:
             namen2_to_inc = gp.add_var_names(
-                v_dict['id_name'], v_dict['cluster_name'], v_dict['w_name'],
+                v_dict['id_name'], v_dict['w_name'],
                 v_dict['x_balance_name'], v_dict['x_name'])
             if c_dict['atet_flag']:
                 namen2_to_inc.extend(v_dict['d_name'])
@@ -241,13 +280,14 @@ def ModifiedCausalForest(
                     mcf_data.adjust_variables(v_dict, var_x_type, var_x_values,
                                               c_dict, x_variables_out))
         time11 = time.time()
+        lc_forest = None
         if c_dict['l_centering'] and c_dict['l_centering_new_sample']:
             l_cent_sample, nonlc_sample, _ = gp.sample_split_2_3(
                 indata2, c_dict['lc_sample'], c_dict['l_centering_share'],
                 c_dict['nonlc_sample'], 1-c_dict['l_centering_share'],
                 random_seed=_seed_sample_split,
                 with_output=c_dict['with_output'])
-            (indata2, old_y_name, new_y_name
+            (indata2, old_y_name, new_y_name, lc_forest
              ) = mcf_lc.local_centering_new_sample(
                 l_cent_sample, nonlc_sample, v_dict, var_x_type, c_dict)
             v_dict = mcf_data.adjust_y_names(v_dict, old_y_name, new_y_name,
@@ -264,11 +304,11 @@ def ModifiedCausalForest(
             _seed_sample_split, c_dict['with_output'])
         if c_dict['l_centering'] and not c_dict['l_centering_new_sample']:
             time11 = time.time()
-            if c_dict['fs_yes']:
+            if c_dict['fs_yes'] and c_dict['fs_other_sample']:
                 files_to_center = (tree_sample, fill_y_sample, fs_sample)
             else:
                 files_to_center = (tree_sample, fill_y_sample)
-            old_y_name, new_y_name = mcf_lc.local_centering_cv(
+            old_y_name, new_y_name, lc_forest = mcf_lc.local_centering_cv(
                 files_to_center, v_dict, var_x_type, c_dict)
             time12 = time.time()
             v_dict = mcf_data.adjust_y_names(v_dict, old_y_name, new_y_name,
@@ -282,7 +322,6 @@ def ModifiedCausalForest(
                       'is activated. Therefore, the part of the input that is',
                       'used for estimating the forest is not used as',
                       'reference sample to get more reliable inference')
-
 # Descriptive statistics by treatment for outcomes and balancing variables
         if c_dict['with_output'] and c_dict['desc_stat']:
             variables_to_desc = [*v_dict['y_name'], *v_dict['x_balance_name']]
@@ -293,7 +332,6 @@ def ModifiedCausalForest(
                 gp_mcf.statistics_by_treatment(indata2, v_dict['d_name'],
                                                variables_to_desc)
             mcf_data.variable_features(var_x_type, var_x_values)
-
 # Common support
     if c_dict['common_support'] > 0:
         if not c_dict['train_mcf']:
@@ -309,6 +347,47 @@ def ModifiedCausalForest(
             d_train_tree)
     else:
         preddata3 = preddata2
+        prob_score, d_train_tree = None, None
+    if c_dict['reduce_training'] or c_dict['reduce_largest_group_train']:
+        if c_dict['fs_yes']:
+            infiles = (tree_sample, fill_y_sample, fs_sample)
+        else:
+            infiles = (tree_sample, fill_y_sample)
+        if c_dict['with_output']:
+            print('=' * 80)
+            for file in infiles:
+                print(file)
+                gp_mcf.statistics_by_treatment(file, v_dict['d_name'],
+                                               v_dict['y_name'])
+        if c_dict['reduce_training']:
+            if c_dict['with_output']:
+                print('Randomly deleting training data.')
+            mcf_data.random_obs_reductions(
+                infiles, fraction=c_dict['reduce_training_share'])
+        if c_dict['reduce_largest_group_train']:
+            if c_dict['with_output']:
+                print('Randomly deleting training data of largest treatment.')
+            mcf_data.random_obs_reductions_treatment(
+                infiles, fraction=c_dict['reduce_largest_group_train_share'],
+                d_name=v_dict['d_name'])
+        if c_dict['with_output']:
+            print('=' * 80)
+            for file in infiles:
+                print(file)
+                gp_mcf.statistics_by_treatment(file, v_dict['d_name'],
+                                               v_dict['y_name'])
+    if c_dict['reduce_prediction']:
+        if c_dict['with_output']:
+            print('=' * 80)
+            print('Randomly deleting prediction data.')
+        infiles = (preddata3,)
+        outfile = (preddata3[:-4] + 'red.csv',)
+        mcf_data.random_obs_reductions(
+            infiles, outfiles=outfile,
+            fraction=c_dict['reduce_prediction_share'])
+        preddata3 = outfile[0]
+        if c_dict['with_output']:
+            print('-' * 80)
 
 # Pre-analysis feature selection
     if c_dict['train_mcf']:
@@ -321,11 +400,11 @@ def ModifiedCausalForest(
                 gp.statistics_covariates(fs_in, var_x_type)
                 print('\n\nFeature selection')
             cfs_dict = copy.deepcopy(c_dict)
-            if not isinstance(cfs_dict['grid_n_min'], int):
+            if isinstance(cfs_dict['grid_n_min'], (list, tuple)):
                 cfs_dict['grid_n_min'] = cfs_dict['grid_n_min'][0]
             cfs_dict['m_grid'] = 1
             # cfs_dict['alpha_reg_grid'] = 1
-            if not isinstance(cfs_dict['grid_alpha_reg'], int):
+            if isinstance(cfs_dict['grid_alpha_reg'], (list, tuple)):
                 cfs_dict['grid_alpha_reg'] = cfs_dict['grid_alpha_reg'][0]
             fs_f, x_name_mcf = mcf_forest.build_forest(
                 fs_in, var_fs, var_x_type, var_x_values, cfs_dict)
@@ -372,14 +451,7 @@ def ModifiedCausalForest(
                 c_dict, x_name_mcf)
         else:
             raise Exception('Currently deactivated. Dics not saved.')
-            # time4 = time.time()
-            # time5 = time.time()
-            # loaded_tuple = gp.save_load(c_dict['save_forest_file_pickle'],
-            #                             save=False,
-            #                             output=c_dict['with_output'])
-            # forest = loaded_tuple[0]
-            # v_dict = loaded_tuple[7]
-            # x_name_mcf = loaded_tuple[8]
+
         time6 = time.time()    # Forest is tuple
         if c_dict['with_output'] and c_dict['verbose']:
             print()
@@ -389,28 +461,8 @@ def ModifiedCausalForest(
                 flush=True)
             print('-' * 80)
     else:
-        time11 = time.time()
-        time12 = time.time()
-        time2 = time.time()
-        time3 = time.time()
-        time4 = time.time()
-        time5 = time.time()
-        time6 = time.time()
-# # Common support
-#     if c_dict['common_support'] > 0:
-#         if not c_dict['train_mcf']:
-#             indatei_tree = None
-#             prob_score = np.load(c_dict['save_forest_file_ps'])
-#             d_train_tree = np.load(c_dict['save_forest_file_d_train_tree'])
-#         else:
-#             prob_score = None
-#             d_train_tree = None
-#         (preddata3, indatei_tree, fill_y_sample, common_support_list,
-#          prob_score, d_train_tree) = mcf_cs.common_support(
-#           indatei_tree, fill_y_sample, preddata2, var_x_type, v_dict, c_dict,
-#             common_support_list, prime_values_dict, prob_score, d_train_tree)
-#     else:
-#         preddata3 = preddata2
+        time11 = time12 = time2 = time3 = time4 = time5 = time6 = time.time()
+
     if c_dict['save_forest'] and c_dict['train_mcf']:
         save_train_data_for_pred(fill_y_sample, v_dict, c_dict, prob_score,
                                  d_train_tree)
@@ -419,7 +471,7 @@ def ModifiedCausalForest(
             (forest, x_name_mcf, c_dict['max_cats_z_vars'], d_in_values,
              no_val_dict, q_inv_dict, q_inv_cr_dict, prime_values_dict,
              unique_val_dict, common_support_list, z_new_name_dict,
-             z_new_dic_dict, c_dict['max_cats_cont_vars']),
+             z_new_dic_dict, c_dict['max_cats_cont_vars'], lc_forest),
             save=True, output=c_dict['with_output'])
     del prob_score, d_train_tree, common_support_list, unique_val_dict
     del z_new_name_dict, d_in_values, no_val_dict, q_inv_dict, q_inv_cr_dict
@@ -440,13 +492,9 @@ def ModifiedCausalForest(
             print('-' * 80)
         if not (c_dict['marg_plots'] and c_dict['with_output']):
             del forest
+
         # Estimation and inference given weights
         time8 = time.time()
-        # time_string = [    'Weight computation:              '
-        #                 ]
-        # time_difference = [time8 - time7]
-        # gp.print_timing(time_string, time_difference)  # print to file
-        # sys.exit()
         w_ate, _, _, ate, ate_se, effect_list = mcf_ate.ate_est(
             weights, preddata3, y_train, cl_train, w_train, v_dict, c_dict)
         time9_ate = time.time()
@@ -456,20 +504,28 @@ def ModifiedCausalForest(
                 var_x_type, var_x_values, w_ate)
             del forest
         time9_marg = time.time()
-        if c_dict['with_output']:
-            mcf_gate.gate_est(weights, preddata3, y_train, cl_train, w_train,
-                              v_dict, c_dict, var_x_type, var_x_values, w_ate,
-                              ate, ate_se)
+        # if c_dict['with_output']:
+        if c_dict['gate_yes']:
+            gate, gate_se = mcf_gate.gate_est(
+                weights, preddata3, y_train, cl_train, w_train, v_dict, c_dict,
+                var_x_type, var_x_values, w_ate, ate, ate_se)
+        else:
+            gate = gate_se = None
         time9_gate = time.time()
-        pred_outfile, _, _, _, _, names_pot_iate = mcf_iate.iate_est_mp(
-            weights, preddata3, y_train, cl_train, w_train, v_dict, c_dict,
-            w_ate)
+        if c_dict['iate_flag']:
+            (pred_outfile, pot_y, pot_y_var, iate, iate_se, names_pot_iate
+             ) = mcf_iate.iate_est_mp(
+                weights, preddata3, y_train, cl_train, w_train, v_dict, c_dict,
+                w_ate, lc_forest=None)
+        else:
+            pot_y = pot_y_var = iate = iate_se = None
+            pred_outfile = names_pot_iate = None
         del _, w_ate
         time9_iate = time.time()
 
         if c_dict['with_output'] and c_dict['balancing_test_w']:
             mcf_ate.ate_est(weights, preddata3, x_bala_train, cl_train,
-                            w_train, v_dict, c_dict, True)
+                            w_train, v_dict, c_dict, True, None)
         del weights, y_train, x_bala_train, cl_train, w_train
         time9_bal = time.time()
 
@@ -479,14 +535,8 @@ def ModifiedCausalForest(
                 c_dict, var_x_type)
         time10 = time.time()
     else:
-        time7 = time.time()
-        time8 = time.time()
-        time9_ate = time.time()
-        time9_marg = time.time()
-        time9_gate = time.time()
-        time9_iate = time.time()
-        time9_bal = time.time()
-        time10 = time.time()
+        time7 = time8 = time9_ate = time9_marg = time9_gate = time.time()
+        time9_iate = time9_bal = time10 = time.time()
 # Finally, save everything: Con, Var (auch wegen recoding), forests etc.
 
     # Print timing information
@@ -555,6 +605,8 @@ def ModifiedCausalForest(
         else:
             outfiletext.close()
         sys.stdout = orig_stdout
+    return (ate, ate_se, gate, gate_se, iate, iate_se, pot_y, pot_y_var,
+            pred_outfile)
 
 
 def save_train_data_for_pred(data_file, v_dict, c_dict, prob_score,
@@ -565,7 +617,7 @@ def save_train_data_for_pred(data_file, v_dict, c_dict, prob_score,
     if not regrf:
         names_to_save.append(*v_dict['d_name'])
     if c_dict['cluster_std']:
-        names_to_save.append.append(*v_dict['cluster_name'])
+        names_to_save.append(*v_dict['cluster_name'])
     if c_dict['w_yes']:
         names_to_save.append(*v_dict['w_name'])
     if not regrf:
